/* -*- mode:prolog; -*-
 ******************************************************************
 *
 * ----------------------------------------------------------------
 *
 *  chart.pl -- une implementation d'un algorithme a charte
 *
 * ----------------------------------------------------------------
 * Description
 *    une implementation d'un algorithme a charte de type Earley
 *   on assume d'abord des productions DCG simples
 *               sans chappements
 *               sans oprateurs de disjonction, entrelacement, ...
 *
 *   Ces extensions peuvent tre traites dans un deuxime temps
 *
 *   La directive :-rec_prolog <Pred>.
 *   est utilise pour rendre non tabul Pred
 *  (le but de l'exercice tant justement d'implanter un mchanisme de tabulation !)
 *
 * ----------------------------------------------------------------
 */

:-require 'format.pl'.

:-extensional 'C'/3.

'C'([T|R],T,R).

%% Pour acceder directement aux productions de la grammaire
:-extensional (-->)/2.

%% Pour activer un mode verbose avec des message de trace
:-extensional verbose_opt.

:-rec_prolog verbose/2.
%% Affiche un message si le mode verbose est activ
verbose(Msg,Args) :-
   argv(L),
   ( verbose_opt ->
       format(Msg,Args)
   ;   
       true
   ).

:-rec_prolog earley_parse/3.
%% earley_parse(Goal,L,R)
%% ----------
%% Le point de dpart dans l'algorithme pour l'analyse de Goal entre L et R
%%     - initialise l'agenda
%%     - process l'agenda jusqu'a ce qu'il soit vide
%%     - test la presence d'un item reponse pour Goal entre 

:-rec_prolog process/1.
%% process(Agenda)
%% -----------
%% 1 cycle de traitement:
%%     - si l'agenda n'est pas vide, prendre le 1er item et le traiter

:-rec_prolog process_item/3.
%% process_item(Item,OldAgenda,Agenda)
%% --------
%% Traitement de Item, avec empilement ventuel de nouveaux items
%% Les items correspondent  des constituants
%%           completement reconnus:  (A --> true)(I,J)
%%           partielement reconnus: (A --> B,Beta)(I,J)

:-rec_prolog predict/3.
%% predict(A(I),OldAgenda,Agenda)
%% ----------
%% Collecte tous les items de prdictions,
%% pour un non terminal A  partir de la position I

:-rec_prolog prediction/2.
%% predication(A(I),NewItem)
%% --------
%% Prediction d'un item de prdiction non dj tabul 
%% pour un non terminal A  partir de la position I
%% en utilisant les production de la grammaire

%% NOTE: on utilisira dcg_normalize(Body,XBody)
%% pour normaliser le corps des productions

:-rec_prolog resolve_passive/3.
%% resolve_passive(B(J,K),OldAgenda,Agenda)
%% ----------
%% Collecte tous les items rsultant de la combinaison
%% avec un non terminal B totalement reconnu entre J et K

:-rec_prolog passive/2.
%% passive(B(J,K),NewItem)
%% ----------
%% Construction d'un item non dj tabul
%% par combinaison d'un non terminal B reconnu entre J et K
%% avec un item en attente de B

:-std_prolog resolve_active/3.
%% resolve_active(Item,OldAgenda,Agenda)
%% ----------
%% Collecte tous les items rsultant de l'utilisation d'un
%% item en attente d'un non terminal B ou d'un terminal [Token]

:-rec_prolog active/2.
%% active(Item,NewItem)
%% ----------
%% Construction d'un item NewItem non dj tabul
%% par combinaison d'un item Item en attente 
%% avec un item en attente de B d'un non terminal B ou d'un terminal [Token]

:-rec_prolog scan/3.
%% scan(Tokens,L,R)
%% Lit une liste de terminaux Tokens sur la chane d'entre

:-rec_prolog stored/1.
%% stored(Item)
%% ----------
%% Test si Item est tabul
%% NOTE: on peut utiliser le prdicat builtin DyALog recorded(Object)

:-rec_prolog store/1.
%% store(Item)
%% -------------
%% Tabule Item si non dj tabul
%% echec si dj tabul
%% NOTE: on peut utiliser le prdicat builtin DyALog record(Object)

:-rec_prolog dcg_normalize/2.
%% dcg_normalize(Body,XBody)
%% normalise les corps de production pour leur ajouter un 'true' final
%% de manire  rendre plus uniforme les traitements
dcg_normalize(Body,XBody) :-
	( Body = (B,Rest) ->
	    dcg_normalize(Rest,XRest),
	    XBody = (B,XRest)
	;
	    XBody = (Body,true)
	).

%% all_solutions(NewItem,Item^Goal,Agenda1,Agenda2)
%% ---------
%% Ajoute  Agenda1
%% l'ensemble des nouveaux items NewItem produits
%% par application de Goal sur l'item slectionn Item

%% NOTE: xcompiler tend le compilateur par expansion de termes
%% toute occurence de all_solutions(...) est remplac (inlining) par le corps
%% de la dfinition de all_solutions

:-xcompiler((all_solutions(NewItem,Item^Goal,Agenda1,Agenda2)
	    :-
	    mutable(M,Agenda1),
	     every((Goal,
		    mutable_read(M,_A),
		    mutable(M,[NewItem|_A]))),
	     mutable_read(M,Agenda2),
	     true
	    )).

:-rec_prolog get_sentence/1.
get_sentence(Sentence) :-
	argv(L),
	( L=['-verbose'|Sentence] ->
	    record(verbose_opt)
	;
	    L = Sentence
	).

%% La requte,
%% ./chart french.dcg -- Paul mange une pomme
%% ./chart french.dcg -- -verbose Paul mange une pomme
?- get_sentence(Sentence),
   earley_parse(s,Sentence,[]).

